#include <windows.h>
#include <stdio.h>

#define CHUNKSIZE (1024*16) //16KiByte
#define MEGA (1024*1024)
#define MAX_MEM  (MEGA*1024) // 1GiByte


char *nextchunk, *curr, *pEnd;
LPVOID pBase;

void printErrorAndExit(const char *msg, DWORD err)
{
	LPSTR lpMsgBuf;
	if(FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
			          FORMAT_MESSAGE_FROM_SYSTEM | 
					  FORMAT_MESSAGE_IGNORE_INSERTS,
						NULL,
						err,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					  (LPTSTR) &lpMsgBuf,
					  0,
					  NULL ))
	{
		fprintf(stderr,"%s : %s\n",msg,lpMsgBuf);
		LocalFree(lpMsgBuf);
	}
	else
	{
		fprintf(stderr,"Error at FormatMesage: %d\n",err=GetLastError());
	}
	exit(err);
}

int main(int argc, char *argv[])
{
	HANDLE hOut,hIn;
	DWORD read=0,cnt,written;
	
	/* check arguments */
	if(argc<3)
	{
		printf("usage: %s file_to_invert result\n",argv[0]);
		exit(-1);
	}
	/* open file for reading and writting */
	hIn=CreateFile(argv[1],GENERIC_WRITE|GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(INVALID_HANDLE_VALUE==hIn)
	{
		printErrorAndExit("Error at CreateFile",GetLastError());
	}

	hOut=CreateFile(argv[2],GENERIC_WRITE|GENERIC_READ,0,NULL,/*CREATE_NEW*/CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	if(INVALID_HANDLE_VALUE==hOut)
	{
		printErrorAndExit("Error at CreateFile",GetLastError());
	}

	/* reserve 1GByte virtual memory , no backing store*/
	pBase=VirtualAlloc(NULL,MAX_MEM,MEM_RESERVE,PAGE_NOACCESS);
	if(NULL==pBase)
	{
		printErrorAndExit("Error at VirtualAlloc reserve",GetLastError());
	}

	curr=nextchunk=(char*)pBase;
	
	do
	{	/* read bytes */
	    curr+=read;
		/* commited memory available */
		if(curr==nextchunk)
		{
			/* commit next chunk */
			if(NULL==VirtualAlloc(nextchunk,CHUNKSIZE,MEM_COMMIT,PAGE_READWRITE))
			{
				printErrorAndExit("VirtualAlloc commit",GetLastError());
			}
			nextchunk+=CHUNKSIZE;
		}
				
	/* read data */
	}while(ReadFile(hIn,curr,nextchunk-curr,&read,NULL) && read);

    /* one char back, last read char */
	curr-=1;
	pEnd=curr;
    if(*curr!='\n')/*invert is impossible if lines not terminates with \n*/
    {
        fprintf(stderr,"warning: no newline at end of file\n");
    }

	while( curr != (char*)pBase )
	{            
		
		if( *curr=='\n' )
		{
			cnt=pEnd-curr;
			if(!WriteFile(hOut,curr+1,cnt,&written,NULL))
			{
				printErrorAndExit("Error at WriteFile reserve",GetLastError());
			}
			pEnd=curr;
		}
        curr--;
	}
	
	/* write remaining bytes */
	if(!WriteFile(hOut,curr,(pEnd-curr)+1,&written,NULL))
	{
		printErrorAndExit("Error at WriteFile reserve",GetLastError());
	}

	if(!VirtualFree(pBase,0,MEM_RELEASE))
	{
		printErrorAndExit("Error at VirtualFree release",GetLastError());
	}

	/* cleanup handles */
	if(!CloseHandle(hIn) || !CloseHandle(hOut) )
	{
		printErrorAndExit("Error at CloseHandle",GetLastError());
	}
	return 0;
}